/* chardev.c 
 *  Simple character device driver to implement a non-blocking
 *    character FIFO.
 *  Derived from experiment one.
 *
 *
 *
 *  When you load this module you will get instructions for
 *    creating the appropriate device file with the kernel-assigned
 *    major device number (probably 254). For example,
 *    mknod ./dev/mydev c 254 0  (do man on mknod).
 *  To invoke this module after creating the device file dev/mydev
 *    do a command line "head dev/mydev" to see
 *    the open and read counts.
 */


/* kernel module programming */
#ifndef MODULE
#define MODULE
#endif

#ifndef LINUX
#define LINUX
#endif

#ifndef __KERNEL__
#define __KERNEL__
#endif

#include <linux/module.h>    /* needed by all modules */
#include <linux/kernel.h>    /* for KERN_ALERT */
#include <linux/fs.h>        /* for struct file_operations */
#include <asm/uaccess.h>     /* for put_user */
#include <asm/errno.h>       /* for all those handy error constants */


/* prototypes - this would normally go in a header file */
static int device_open(struct inode *, struct file *);
static int device_release(struct inode *, struct file *);
static ssize_t device_read(struct file *, char *, size_t, loff_t *);
static ssize_t device_write(struct file *, const char *, size_t, loff_t *);
int init_module(void);
void cleanup_module(void);

#define DEVICE_NAME "mydev"    /* device name in /proc/devices */
#define MAX_MSG_BUF 81         /* max user message buffer size plus 0 EOS */
#define MAX_DEV_BUF 256        /* max device buffer size */

/* Device driver global variables */
static int Major;
static int dev_open = 0;

static int wp = 0;          /* write pointer */
static int rp = 0;          /* read pointer chasing the write pointer */

static char devbuf[MAX_DEV_BUF]


/* Gcc assigns the rest of the structure fields to NULL */
static struct file_operations fops = {
  .read = device_read,
  .write = device_write,
  .open = device_open,
  .release = device_release
};


/* new character device file functions */

static int device_open(struct inode *inodp, struct file *filp)
{ 
  /* This seems like it is not concurrent-safe locking */
  if (dev_open) return -EBUSY;

  dev_open++;

  wp = 0;
  rp = 0;

  MOD_INC_USE_COUNT;

  return 0;
}

static int device_release(struct inode *inodp, struct file *filp)
{

  dev_open--;

  MOD_DEC_USE_COUNT;

  return 0;
}

static ssize_t device_read(struct file *filp,
  char *buf, 		/* caller's buffer  and length */
  size_t length,
  loff_t *offset)	/* Our offset in the file */
{ static char msg[MAX_BUF];
  int len;

  /* Figure out when and how to send back an eof (use
   * a return 0 to signify no bytes read to the file I/O
   * driver).
   * Programs like cat won't quit until they see an eof.
   * Can I use offset pointer to do this? Say after a single read?
   * if (offset > 0) return 0;
   */


  /* use put_user to copy the devbuf to the user buf from the current
   * wp up to a 0 string terminator if the device buffer is not
   * empty (wp == rp)
   *
   * We won't let the user write more than 80 characters into the
   * the device buffer at a time.
   */


  /* len back from sprintf does not include the trailing 0 */
  if ( copy_to_user(buf, msg, len+1) )
    return -EFAULT;

  return len;

}
  
static ssize_t device_write(struct file *filp,
  const char *buf,   /* caller's buffer  and length */
  size_t length,
  loff_t *offset)    /* Our offset in the file */
{
  /* use get_user or copy_from_user to copy the msg buffer
   * into the device buffer up to a maximum of 80. I think
   * we can just read to the 0 EOS, and if we get to 80
   * then jam a 0 EOS. Also we must not read past 
   * a full device buffer (wp + 1 == rp). 
   */ 

}



/* Initialize the module at insmod */
int init_module(void)
{ 
  printk(KERN_ALERT "chardev module loaded\n");

  /* the first argument as 0 directs kernel to supply a Major device number */
  Major = register_chrdev(0, DEVICE_NAME, &fops);

  if (Major < 0)
  { printk (KERN_ALERT "init_module: register char dev failed with %d\n", Major);
    return Major;
  }

  printk(KERN_ALERT "init_module: Major device number assigned is: %d\n", Major);  
  printk(KERN_ALERT "init_module: Create device with mknod ./dev/mydev c %d 0\n", Major); 

  /* A non-zero return means init_module failed; module can't be loaded */
  return 0;
}


/* Clean up the module at rmmod */
void cleanup_module(void)
{ int ret;

  ret = unregister_chrdev(Major, DEVICE_NAME);
  if (ret < 0) printk(KERN_ALERT "cleanup_module: unregister failed with %d\n", ret);

  printk(KERN_ALERT "chardev module unloaded\n");
}

MODULE_LICENSE("GPL");
